unit mQuery;

{$I mODBC.INC}

interface

uses
  SysUtils, Classes, Forms, Dialogs, Db,
  mDataBas, odbcsql, mParams;

{$IFDEF DELPHI4}
const
   dsMaxBufferCount = MAXINT DIV 8;
type
  PBufferList = ^TBufferList;
  TBufferList = array[0..dsMaxBufferCount - 1] of PChar;
{$ENDIF}
type
  PWord = ^Word;
  PByte     =^Byte;
  TmCustomQuery = class;
{ TBlobStream }

  TmBlobStream = class(TStream)
  private
    FField:    TBlobField;
    FDataSet:  TmCustomQuery;
    FBuffer:   PChar;            //    
    FMode:     TBlobStreamMode;
    FFieldNo:  Integer;
    FOpened:   Boolean;
    FModified: Boolean;
    FPosition: Longint;
    function GetBlobSize: Longint;
  public
    constructor Create(Field: TBlobField; Mode: TBlobStreamMode);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    procedure Truncate;
  end;

  TODBCFieldDef = class(TFieldDef)
  private
     FOffset:      integer;
     FSQLDataType: SQLSMALLINT;
     FSQLsize:     SQLUINTEGER;
  public
//     BaseTable,
//     BaseColumn:   String;
     property OffsetInbuf: integer read fOffset write Foffset;
     property SQLDataType: SQLSMALLINT read fSQLDataType write FSQLDataType;
     property SQLsize:     SQLUINTEGER read fSQLsize write FSQLsize;
  end;

  TODBCRecInfo = record
    RecordNumber: SQLINTEGER;
    UpdateStatus: TUpdateStatus;
    BookmarkFlag: TBookmarkFlag;
  end;
  PODBCRecInfo = ^TODBCRecInfo;

  PBookMarkList = ^TBookMarkList;
  TBookMarkList = array[0..dsMaxBufferCount - 1] of Char;
  TmCursorTypes = (ctFwdOnlyCursor,ctStaticCursor,ctKeySetCursor,ctDynamicCursor);
  TmCursorAttributes = record
    Attrib1:           SQLUINTEGER;
    Attrib2:           SQLUINTEGER;
    UseDriverBookMark: Boolean;
    Concurrency:       SQLUINTEGER;
    Sensitivity:       SQLUINTEGER;
  end;

  TmQueryStateBit = (mqOpened, mqPrepared, mqInitialized);
  TmQueryState = set of TmQueryStateBit;

  { TQueryDataLink }

  TmCustomQueryDataLink = class(TDataLink)
  private
    FQuery: TmCustomQuery;
  protected
    procedure ActiveChanged; override;
    procedure RecordChanged(Field: TField); override;
    procedure CheckBrowseMode; override;
  public
    constructor Create(AQuery: TmCustomQuery);
  end;

  TmCustomQuery = class(TDataSet)
  private
    { Private declarations }
    hstmt:         SQLHSTMT;
    FetchStatus:   TBookMarkFlag;

    FBlobCount:    Integer;
    FCanModify:    boolean;
//    FCacheBlobs:   Boolean;
    FCachedUpdates:Boolean;
    FCursorAttr:   TmCursorAttributes; // ODBC attributes of SQL cursor
    FCursorType:   TmCursorTypes;      // type cursor (static,dynamic)
    FDataBase:     TmDataBase;
    FDeletedCount: Integer;            // count of deleted records
    FDeletedRec:   PBookMarkList;      // bookmarks of deleted records
    FCachedCount:  Integer;            // count of cached records
    FCachedBuffer: PBufferList;        // buffers with cached records
    FParamCheck:   Boolean;
    FParams:       TmParams;
    FRecordSize:   Word;
    FRecBufSize:   Word;
    FRecInfoOfs:   Word;
    FBlobCacheOfs: Word;
    FRequestLive:  boolean;
    FRowsAffected: LongInt;
    FSQL:          TStrings;
    FState:        TmQueryState;
    FupdateSQL:    array[TUpdateKind] of TStrings;
    FText:         String;
    FDataLink:     TDataLink;
    FQueries:      array[TUpdateKind] of TmCustomQuery;
    FoldBuffer:    PChar; // buffer for old values of edited record
    FBindBuffer:   PChar; // buffer for SQLbindColumn
    FFilterBuffer: PChar;
    FonFilter:     TFilterRecordEvent;
    procedure AddDeletedToList(Buffer: Pointer);
    procedure AddToCacheBuffer( uStatus:TUpdateStatus; Buffer: PChar);
    function  CheckDeleted( Buffer: Pointer):boolean;
    procedure CheckSQLResult( sqlres:SQLRETURN);
    procedure CreateParams(List: TmParams; const Value: PChar);
    procedure InitStmt;
    procedure FreeDeletedList;
    procedure FreeModifiedList;
    function  GetActiveRecBuf(var RecBuf: PChar): Boolean;
    function  GetBlobData(Field: TField; Buffer: PChar): TmBlobData;
    function  GetCursorName:String;
//    function  GetRowsAffected: LongInt;
    procedure GetStmtAttributes;
    procedure QueryChanged(Sender: TObject);
    procedure RefreshParams;
    procedure SetBlobData(Field: TField; Buffer: PChar; Value: TmBlobData);
    procedure SetDataBase(Value: TmDataBase);
    procedure SetDataSource(Value: TDataSource);
    procedure SetParamsFromCursor;
    procedure SetParamsList(Value: TmParams);
    function  getBookmarkPtr(Buffer: Pointer):Pointer;
    procedure InternalOpenCursor;
    procedure CopyRecordBuffer( BufferSrc, BufferDst:PChar);
    function  GetUpdateSQL(Index: Integer): TStrings;
    function  GetUpdateQuery(UpdateKind: TUpdateKind): TmCustomQuery;
    procedure SQLBindColumns( Buffer: PChar);
    procedure SQLUnBindColumns;
    function  CheckFilter( Buffer: Pointer): boolean;
    function  GetPrepared: Boolean;
//    procedure SetPrepared(const Value: Boolean);
    { std }
  protected
    { Protected declarations }
    function AllocRecordBuffer: PChar; override;
    procedure ClearCalcFields(Buffer: PChar); override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function  GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function  GetDataSource: TDataSource; override;
    function GetRecordSize: Word; override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetCachedUpdates(Value: Boolean);
{$IFNDEF DELPHI4}
    function  GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
{$ENDIF}
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    procedure SetCursorType(Value: TmCursorTypes);
    function  GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function  GetRecNo: Integer; override;
    function  GetRecordCount: Integer; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalEdit; override;
    procedure InternalPost; override;
    procedure PrepareCursor; virtual;
    procedure InternalCancel; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    procedure InternalRefresh; override;
    function IsCursorOpen: Boolean; override;
    function GetCanModify: Boolean; override;
    procedure SetQuery(Value: TStrings); virtual;
    procedure SetUpdateSQL(Index: Integer; Value: TStrings); virtual;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure UnprepareCursor; virtual;

    property CachedUpdates: Boolean read FCachedUpdates write SetCachedUpdates default False;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure ApplyUpdates;
    procedure CancelUpdates;
    procedure CommitUpdates;
    function  CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
{$IFDEF DELPHI4}
    function  GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
{$ENDIF}
    property  CursorAttr: TmCursorAttributes read FCursorAttr;
    property  CursorName:String read GetCursorName;
    procedure ExecSQL;
    procedure FreeStmt;
    function  IsSequenced: Boolean; override;
    function  Locate( const KeyFields: string; const KeyValues: Variant;
                      Options: TLocateOptions): Boolean; override;
    function  Lookup( const KeyFields: string; const KeyValues: Variant;
                      const ResultFields: string): Variant; override;
    function  ParamByName(const Value: string): TmParam;
    procedure SetParamsForUpdateSql(UpdateKind: TUpdateKind);
    property  RowsAffected: longint read FRowsAffected;
    property  Text: string read FText;
    property  SQL: TStrings read FSQL write SetQuery;
    property  ModifySQL: TStrings index 0 read GetUpdateSQL write SetUpdateSQL;
    property  InsertSQL: TStrings index 1 read GetUpdateSQL write SetUpdateSQL;
    property  DeleteSQL: TStrings index 2 read GetUpdateSQL write SetUpdateSQL;
    property  DataBase:TmDataBase read FDataBase write SetDataBase;
    property  DataSource: TDataSource read GetDataSource write SetDataSource;
    property  RequestLive:boolean read FRequestLive write FRequestLive default False;
    property  CursorType:TmCursorTypes read FCursorType write SetCursorType default ctDynamicCursor;
    property  ParamCheck: Boolean read FParamCheck write FParamCheck default True;
    property  Params: TmParams read FParams write SetParamsList;
    property  Prepared: Boolean read GetPrepared;
    property  OnFilterRecord: TFilterRecordEvent read FOnFilter write FOnFilter;
  published
    { Published declarations }
  end;

  TmQuery = class(TmCustomQuery)
  published
    { Published declarations }
    property Active;
    property AutoCalcFields;
    property SQL;
    property ModifySQL;
    property InsertSQL;
    property DeleteSQL;
    property DataBase;
    property DataSource;
    property RequestLive;
    property CursorType;
    property ParamCheck;
    property Params;
    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;
  end;


implementation
uses DbConsts, mConst, mExcept;

  function MemCompare( Buf1,Buf2: PChar; Count:Integer):boolean;
  Var
     i: Integer;
  begin
       Result:=False;
       for i:=0 to Count-1 do
         if Buf1[i]<>Buf2[i] then Exit;
       Result:=True;
  end;

{ TQueryDataLink }

constructor TmCustomQueryDataLink.Create(AQuery: TmCustomQuery);
begin
  inherited Create;
  FQuery := AQuery;
end;

procedure TmCustomQueryDataLink.ActiveChanged;
begin
  if FQuery.Active then FQuery.RefreshParams;
end;

procedure TmCustomQueryDataLink.RecordChanged(Field: TField);
begin
  if (Field = nil) and FQuery.Active then FQuery.RefreshParams;
end;

procedure TmCustomQueryDataLink.CheckBrowseMode;
begin
  if FQuery.Active then FQuery.CheckBrowseMode;
end;

{ TmCustomQuery }

procedure TmCustomQuery.CheckSQLResult( sqlres:SQLRETURN);
begin
   case sqlres of
     SQL_SUCCESS:;
     SQL_SUCCESS_WITH_INFO: raise ESQLerror.CreateDiag(SQL_HANDLE_STMT,hstmt, sqlres);
     SQL_NEED_DATA:         raise ESQLerror.Create('SQL_NEED_DATA');
     SQL_STILL_EXECUTING:   raise ESQLerror.Create('SQL_STILL_EXECUTING');
     SQL_ERROR:             raise ESQLerror.CreateDiag(SQL_HANDLE_STMT,hstmt, sqlres);
     SQL_NO_DATA:           raise ESQLerror.Create('SQL_NO_DATA');
     SQL_INVALID_HANDLE:    raise ESQLerror.Create('SQL_INVALID_HANDLE');
     else                   raise ESQLerror.Create('unknown SQL result');
     end;
end;

constructor TmCustomQuery.Create(AOwner: TComponent);
Var
   UpdateKind: TUpdateKind;
begin
  inherited Create(AOwner);
  FSQL           := TStringList.Create;
  hstmt          := 0;
  FRecBufSize    := 0;
  FDeletedCount  := 0;
  FCanModify     := False;
  FRequestLive   := False;
  ParamCheck     := True;
  FCachedCount   := 0;
  FCachedBuffer  := nil;
  FParams        := TmParams.Create;
  FDataLink      := TmCustomQueryDataLink.Create(Self);
  TStringList(SQL).OnChange := QueryChanged;
  FCursorType    := ctDynamicCursor;
  FState         := [];

  FCachedUpdates := False;
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
    FupdateSQL[UpdateKind] := TStringList.Create;
  FoldBuffer     := nil;
  FBindBuffer    := nil;
end;

destructor TmCustomQuery.Destroy;
Var
   UpdateKind: TUpdateKind;
begin
  FreeStmt;
  if Assigned(FDataBase) then FDataBase.ExcludeDataSet(Self);
  SQL.Free;
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
    FupdateSQL[UpdateKind].free;
  FreeDeletedList;
  FreeModifiedList;
  FParams.Free;
  FDataLink.free;
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
    if Assigned(FQueries[UpdateKind])
      then FQueries[UpdateKind].free;
  inherited Destroy;
end;

procedure TmCustomQuery.CopyRecordBuffer( BufferSrc, BufferDst: PChar);
Var
   i: LongInt;
begin
     Move( BufferSrc^, BufferDst^, FBlobCacheOfs);
     for i:=0 to FBlobCount-1 do
       PmBlobDataArray(BufferDst + FBlobCacheOfs)[i]:=
            PmBlobDataArray(BufferSrc + FBlobCacheOfs)[i];
     Move( (BufferSrc+FRecInfoOfs)^,
           (BufferDst+FRecInfoOfs)^, SizeOf(TODBCRecInfo));
end;


function TmCustomQuery.getBookmarkPtr(Buffer: Pointer):Pointer;
begin
  if CursorAttr.UseDriverBookMark
    then Result:=Buffer
    else Result:=@(PODBCRecInfo(PChar(Buffer) + FRecInfoOfs).RecordNumber);
end;

{ list of bookmarks deleted records }
procedure TmCustomQuery.FreeDeletedList;
begin
  if FDeletedCount>0 then FreeMem( FDeletedRec);
  FDeletedCount:=0;
end;

procedure TmCustomQuery.AddDeletedToList(Buffer: Pointer);
Var
  NewList: PBookMarkList;
begin
     GetMem(NewList, (FDeletedCount+1) * BookMarkSize);
     if FDeletedCount>0 then
     begin
       Move(FDeletedRec^, NewList^, FDeletedCount * BookMarkSize);
       FreeMem(FDeletedRec);
     end;
     FDeletedRec:=NewList;
     Move( getBookmarkPtr(Buffer)^,
           FDeletedRec^[FDeletedCount*BookMarkSize],
           BookMarkSize);
     inc( FDeletedCount);
end;

procedure TmCustomQuery.AddToCacheBuffer( uStatus:TUpdateStatus; Buffer: PChar);
Var
   i:  LongInt;
   b:  PChar;
begin
    for i:=0 to FCachedCount-1 do
      if (PODBCRecInfo(FCachedBuffer[i] + FRecInfoOfs)^.UpdateStatus = uStatus)
         and MemCompare(getBookmarkPtr( Buffer), getBookmarkPtr( FCachedBuffer[i]), BookMarkSize)
        then begin
           CopyRecordBuffer( Buffer, FCachedBuffer[i]);
           exit;
        end;
    b:=AllocRecordBuffer;
    CopyRecordBuffer( Buffer, b);
    PODBCRecInfo(b + FRecInfoOfs)^.UpdateStatus := uStatus;
    ReallocMem( FCachedBuffer, (FCachedCount+1)*sizeof(PChar));
    FCachedBuffer[FCachedCount]:=b;
    inc( FCachedCount);
end;

procedure TmCustomQuery.FreeModifiedList;
Var
   i: LongInt;
   p: PBufferList;
begin
   // free record buffers
   p:=FCachedBuffer;
   FCachedBuffer:=nil;
   for i:=0 to FCachedCount-1 do begin
      FreeRecordBuffer( p[i]);
   end;
   // free array
   if p<>nil then FreeMem( p, FCachedCount*sizeof(PChar));
   FCachedCount:=0;
end;

function TmCustomQuery.CheckDeleted( Buffer: Pointer):boolean;
Var
   i: Integer;
begin
     Result:=False;
     if (FCursorAttr.Sensitivity and SQL_SS_DELETIONS)=0
     then
        for i:=0 to FDeletedCount-1 do
          if MemCompare(getBookmarkPtr( Buffer), @(FDeletedRec[i*BookMarkSize]), BookMarkSize)
            then Exit;
     Result:=True;
end;

procedure TmCustomQuery.SetDataBase(Value: TmDataBase);
begin
     if FDataBase<>Value then
      begin
         if Assigned(FDataBase) then FDataBase.ExcludeDataSet(Self);
         FDataBase:=Value;
         if Assigned(FDataBase) then FDataBase.IncludeDataSet(Self);
      end;
end;

procedure TmCustomQuery.GetStmtAttributes;
var
   ui: SQLUINTEGER;
begin
     CheckSQLResult( SQLGetStmtAttr( hstmt, SQL_ATTR_CURSOR_TYPE, @ui, sizeof(ui), nil));
     case ui of
              SQL_CURSOR_STATIC:        CursorType:= ctStaticCursor;
              SQL_CURSOR_KEYSET_DRIVEN: CursorType:= ctKeySetCursor;
              SQL_CURSOR_DYNAMIC:       CursorType:= ctDynamicCursor;
              else raise ESQLerror.Create( SmForwardNotSupported);
     end;
     with FCursorAttr do
     begin
          Attrib1:=0; Attrib2:=0;
          Sensitivity:=0;
          try
          case CursorType of
          ctStaticCursor:  begin
              DataBase.ODBCDriverInfo( SQL_STATIC_CURSOR_ATTRIBUTES1, SQLPOINTER(@Attrib1), sizeof(SQLUINTEGER), nil);
              DataBase.ODBCDriverInfo( SQL_STATIC_SENSITIVITY, SQLPOINTER(@Sensitivity), sizeof(SQLUINTEGER), nil);
              if Length(FUpdateSQL[ukModify].Text)>0 // using external update SQL statement
                then Sensitivity:=Sensitivity and (not SQL_SS_UPDATES)
//              DataBase.ODBCDriverInfo( SQL_STATIC_CURSOR_ATTRIBUTES2, SQLPOINTER(@Attrib2), sizeof(SQLUINTEGER), nil);
            end;
          ctKeySetCursor:    begin
              DataBase.ODBCDriverInfo( SQL_KEYSET_CURSOR_ATTRIBUTES1, SQLPOINTER(@Attrib1), sizeof(SQLUINTEGER), nil);
              DataBase.ODBCDriverInfo( SQL_STATIC_SENSITIVITY, SQLPOINTER(@Sensitivity), sizeof(SQLUINTEGER), nil);
              if Length(FUpdateSQL[ukModify].Text)>0 // using external update SQL statement
                then Sensitivity:=Sensitivity and (not SQL_SS_UPDATES)
//              DataBase.ODBCDriverInfo( SQL_KEYSET_CURSOR_ATTRIBUTES2, SQLPOINTER(@Attrib2), sizeof(SQLUINTEGER), nil);
            end;
          ctDynamicCursor:   begin
              DataBase.ODBCDriverInfo( SQL_DYNAMIC_CURSOR_ATTRIBUTES1, SQLPOINTER(@Attrib1), sizeof(SQLUINTEGER), nil);
              Sensitivity:= SQL_SS_ADDITIONS + SQL_SS_DELETIONS + SQL_SS_UPDATES; // by default
//              DataBase.ODBCDriverInfo( SQL_DYNAMIC_CURSOR_ATTRIBUTES2, SQLPOINTER(@Attrib2), sizeof(SQLUINTEGER), nil);
            end;
          end;
          except; end; // some drivers not support second attributes

          DataBase.ODBCDriverInfo( SQL_BOOKMARK_PERSISTENCE, SQLPOINTER(@ui), sizeof(SQLUINTEGER), nil);
          if ((Sensitivity and (SQL_SS_ADDITIONS+SQL_SS_DELETIONS))=0)or(ui=0)
             then UseDriverBookMark:=False
             else UseDriverBookMark:=True;
          if UseDriverBookMark
             then CheckSQLResult( SQLSetStmtAttr( hstmt, SQL_ATTR_USE_BOOKMARKS, Pointer(SQL_UB_VARIABLE{ON}), 0))
             else SQLSetStmtAttr( hstmt, SQL_ATTR_USE_BOOKMARKS, Pointer(SQL_UB_OFF), 0);

          CheckSQLResult( SQLGetStmtAttr( hstmt, SQL_ATTR_CONCURRENCY, @Concurrency, sizeof(Concurrency), nil));
     end;
end;

procedure TmCustomQuery.InitStmt;
Var
   sqlres: SQLRETURN;
   ui:     SQLUINTEGER;
begin
   if (mqInitialized in FState) and
      (hstmt <> 0) then Exit;
   if DataBase=nil then
       raise ESQLerror.Create( SmNoDataSource);
   if DataBase.hdbc=0 then
       DataBase.Connect;
   DataBase.ODBCDriverInfo( SQL_SCROLL_OPTIONS, SQLPOINTER(@ui), sizeof(SQLUINTEGER), nil);
   if ui=SQL_SO_FORWARD_ONLY
          then raise ESQLerror.Create( SmForwardNotSupported);
   sqlres:=SQLAllocHandle(SQL_HANDLE_STMT, DataBase.hdbc, hstmt);
   if sqlres <> SQL_SUCCESS
       then raise ESQLerror.Create( SmAllocateSTMTError);
   Include(FState, mqInitialized);
   try
     with FCursorAttr do
     begin
          try
             case CursorType of
             ctStaticCursor:  CheckSQLResult( SQLSetStmtAttr( hstmt, SQL_ATTR_CURSOR_TYPE, Pointer(SQL_CURSOR_STATIC), SQL_IS_UINTEGER));
             ctKeySetCursor:  CheckSQLResult( SQLSetStmtAttr( hstmt, SQL_ATTR_CURSOR_TYPE, Pointer(SQL_CURSOR_KEYSET_DRIVEN), SQL_IS_UINTEGER));
             ctDynamicCursor: CheckSQLResult( SQLSetStmtAttr( hstmt, SQL_ATTR_CURSOR_TYPE, Pointer(SQL_CURSOR_DYNAMIC), SQL_IS_UINTEGER));
             end;
          except on E: ESQLerror do if (E.SqlState<>'01S02'){and(E.SqlState<>'HYC00')} then raise; end;
          try
             DataBase.ODBCDriverInfo( SQL_SCROLL_CONCURRENCY, SQLPOINTER(@ui), sizeof(SQLUINTEGER), nil);

             if (ui and SQL_SCCO_LOCK)<>0 then CheckSQLResult( SQLSetStmtAttr( hstmt, SQL_ATTR_CONCURRENCY, Pointer(SQL_CONCUR_LOCK), SQL_IS_UINTEGER))
             else if (ui and SQL_SCCO_OPT_ROWVER)<>0 then CheckSQLResult( SQLSetStmtAttr( hstmt, SQL_ATTR_CONCURRENCY, Pointer(SQL_CONCUR_ROWVER), SQL_IS_UINTEGER))
                  else if (ui and SQL_SCCO_OPT_VALUES)<>0 then CheckSQLResult( SQLSetStmtAttr( hstmt, SQL_ATTR_CONCURRENCY, Pointer(SQL_CONCUR_VALUES), SQL_IS_UINTEGER))
                       else if (ui and SQL_SCCO_READ_ONLY)<>0 then CheckSQLResult( SQLSetStmtAttr( hstmt, SQL_ATTR_CONCURRENCY, Pointer(SQL_CONCUR_READ_ONLY), SQL_IS_UINTEGER));
          except on E: ESQLerror do if (E.SqlState<>'01S02'){and(E.SqlState<>'HYC00')} then raise; end;
          GetStmtAttributes;
     end;
   except
     FreeStmt;
     raise;
   end;
end;

procedure TmCustomQuery.FreeStmt;
begin
    if hstmt=0 then exit;
    Close;
    SQLFreeHandle(SQL_HANDLE_STMT,hstmt);
    hstmt:=0;
    FState := [];//Exclude(FState, mqInitialized);
end;

procedure TmCustomQuery.ExecSQL;
   function GetRowsAffected: LongInt;
   Var
      rCount: SQLINTEGER;
   begin
        CheckSQLResult( SQLRowCount( hstmt, rCount));
        Result:= rCount;
   end;
begin
  CheckInActive;
  InternalOpenCursor;
  FRowsAffected:=GetRowsAffected;
  InternalClose;
end;

procedure TmCustomQuery.SetQuery(Value: TStrings);
begin
  if SQL.Text <> Value.Text then
  begin
//    Disconnect;
    SQL.BeginUpdate;
    try
      SQL.Assign(Value);
    finally
      SQL.EndUpdate;
    end;
  end;
end;

procedure TmCustomQuery.SetUpdateSQL(Index: Integer; Value: TStrings);
begin
  if (State <> dsInActive) and (not (csDesigning in ComponentState)) then
    if (Length(FUpdateSQL[ukInsert].Text)>0)<>(Length(Value.Text)>0)
      then raise Exception.Create('You can''t change Post metod while query is active');

  FupdateSQL[TUpdateKind(Index)].Assign(Value)
end;

function TmCustomQuery.GetupdateSQL(Index: Integer): TStrings;
begin
  Result := FupdateSQL[TUpdateKind(Index)];
end;

procedure TmCustomQuery.InternalOpen;
begin
  FetchStatus:=bfCurrent;
  FCanModify := False;
//  FRecNoStatus := TRecNoStatus(CursorProps.ISeqNums);}
  InternalInitFieldDefs;
  if DefaultFields then CreateFields;
  BindFields(True);
  FBlobCacheOfs := FRecordSize + CalcFieldsSize;
  FRecInfoOfs := FBlobCacheOfs + FBlobCount * SizeOf(Pointer);
  FRecBufSize := FRecInfoOfs + SizeOf(TODBCRecInfo);
  //  bookmark     (   )
end;

procedure TmCustomQuery.InternalClose;
begin
     SQLFreeStmt( hstmt, SQL_CLOSE); { SQLCloseCursor( hstmt); }
     Exclude(FState, mqOpened);
     SQLUnBindColumns;

     BindFields(False);
     if DefaultFields then DestroyFields;
     FCanModify := False;
     FreeDeletedList;
     FreeModifiedList;
end;

procedure TmCustomQuery.InternalRefresh;
//Var
//   sqlres: SQLRETURN;
begin
{     SQLCloseCursor( hstmt);
     sqlres:=SQLExecute( hstmt);
     case sqlres of
     SQL_SUCCESS:;
     SQL_SUCCESS_WITH_INFO: raise ESQLerror.CreateDiag(SQL_HANDLE_STMT,hstmt);
     SQL_NEED_DATA:         raise ESQLerror.Create('SQL_NEED_DATA');
     SQL_STILL_EXECUTING:   raise ESQLerror.Create('SQL_STILL_EXECUTING');
     SQL_ERROR:             raise ESQLerror.CreateDiag(SQL_HANDLE_STMT,hstmt);
     SQL_NO_DATA:           raise ESQLerror.Create('SQL_NO_DATA');
     SQL_INVALID_HANDLE:    raise ESQLerror.Create('SQL_INVALID_HANDLE');
     else                   raise ESQLerror.Create('unknown SQL result');
     end;
   if (FCursorAttr.Sensitivity and SQL_SS_DELETIONS)=0
     then FreeDeletedList;
     }
end;

procedure TmCustomQuery.InternalInitFieldDefs;
  procedure AddODBCFieldDesc(FFieldNo:Word);
  var
     FDataType:    TFieldType;
     FSize:        SQLUINTEGER;
     FName:        string;
     pname:       array [0..255] of char;
     nlength:     SQLSMALLINT;
     dtype:       SQLSMALLINT;
     BufDataType: SQLSMALLINT;
     csize:       SQLUINTEGER;
     decdig:      SQLSMALLINT;
     Nullable:    SQLSMALLINT;
     i:           Integer;
{$IFDEF CBUILDER3}
    function AddODBCFieldDef:TODBCFieldDef;
    begin
         Result:=TODBCFieldDef.Create(FieldDefs);
    end;
{$ENDIF}
  begin
    CheckSQLResult( SQLDescribeCol(hstmt,FFieldNo,pname,254,nlength,dtype,csize,decdig,Nullable));
    if FFieldNo=0 then
    begin
         BookmarkSize := cSize;
         FRecordSize  := cSize;
         exit;
    end;
    i := 0;
    if pname[0]=#0 then StrCopy( pname, 'COLUMN');
    FName:=StrPas(pname);
    while FieldDefs.IndexOf(FName) >= 0 do
    begin
         Inc( i );
         FName:=Format('%s_%d', [StrPas(pname), i]);
    end;

    FSize:=cSize;
    BufDataType:=dtype;
    case dtype of
    SQL_CHAR,
    SQL_VARCHAR:    begin FDataType:=ftString; BufDataType:=SQL_CHAR; {inc( Size);} inc( cSize); end;
    SQL_TINYINT,
    SQL_SMALLINT:   begin FDataType:=ftSmallint; BufDataType:=SQL_SMALLINT; FSize:=0; cSize:=sizeof(SQLSMALLINT);end;
    SQL_INTEGER:    begin FDataType:=ftInteger; FSize:=0; cSize:=sizeof(SQLINTEGER);end;
    SQL_DECIMAL,
    SQL_NUMERIC,
    SQL_REAL,
    SQL_FLOAT,
    SQL_DOUBLE:     begin FDataType:=ftFloat; FSize:=0; BufDataType:=SQL_DOUBLE; cSize:=sizeof(SQLDOUBLE);end;
    SQL_DATE,
    SQL_TYPE_DATE:  begin FDataType:=ftDate;  FSize:=0; cSize:=sizeof(SQL_DATE_STRUCT); end;
    SQL_TIME,
    SQL_TYPE_TIME:  begin FDataType:=ftTime;  FSize:=0; cSize:=sizeof(SQL_TIME_STRUCT); end;
    SQL_TYPE_TIMESTAMP,
    SQL_TIMESTAMP:  begin FDataType:=ftDateTime; FSize:=0; cSize:=sizeof(SQL_TIMESTAMP_STRUCT); end;
    SQL_BIT:        begin FDataType:=ftBoolean; FSize:=0; cSize:=sizeof(SQLCHAR); end;
//    SQL_BIGINT:     DataType:=ftUnknown;
    SQL_BINARY,
    SQL_VARBINARY,
    SQL_LONGVARBINARY,
    SQL_LONGVARCHAR:    begin FDataType:=ftMemo; BufDataType:=SQL_BINARY; cSize:=0; FSize := 0; end;
    else FDataType:=ftUnknown;
         FSize:=0;
    end;
    if FDataType <> ftUnknown then
{$IFDEF CBUILDER3}
      with FieldDefs, AddODBCFieldDef do
      begin
        BeginUpdate;
        try
          Name     := FName;
          DataType := FDataType;
          Size     := FSize;
          FieldNo  := FFieldNo;
          Required := (Nullable=SQL_NO_NULLS);
{$ELSE}
      with TODBCFieldDef.Create(FieldDefs, FName, FDataType, FSize, (Nullable=SQL_NO_NULLS), FFieldNo) do
      begin
{$ENDIF}
        if BufDataType<>SQL_BINARY then OffsetInBuf := FRecordSize
                                   else begin OffsetInBuf := FBlobCount; inc(FblobCount);end;
        SQLDataType := BufDataType;
        SQLsize     := cSize;
        FRecordSize:=FRecordSize+cSize+sizeof(SQLINTEGER);
//           InternalCalcField := bCalcField;
        if SQLDataType = SQL_DOUBLE then Precision := decdig;
{$IFDEF CBUILDER3}
        finally
          EndUpdate;
        end;
{$ENDIF}
      end;
  end;
var
   Colmns: SQLSMALLINT;
   i:      integer;
begin
   InternalOpenCursor;
   //   
   FieldDefs.Clear;

   CheckSQLResult( SQLNumResultCols(hstmt, Colmns));
   BookmarkSize := sizeof(SQLUINTEGER);
   FRecordSize  := 0;
   FBlobCount   := 0;
   if CursorAttr.UseDriverBookMark then for i:=0 to Colmns do AddODBCFieldDesc(i)
                                   else for i:=1 to Colmns do AddODBCFieldDesc(i);
   FCanModify:=RequestLive;
end;

procedure TmCustomQuery.PrepareCursor;
begin
    if mqPrepared in FState then Exit;
    CheckSQLResult( SQLPrepare( hstmt, PCHAR(Text), SQL_NTS));
    Include(FState, mqPrepared);
end;

procedure TmCustomQuery.UnPrepareCursor;
begin
  Exclude(FState, mqPrepared);
end;
{
procedure TmCustomQuery.SetPrepared(const Value: Boolean);
begin
  if Value then PrepareCursor
           else UnPrepareCursor;
end;
}
function TmCustomQuery.GetPrepared: Boolean;
begin
  Result := mqPrepared in FState;
end;

procedure TmCustomQuery.InternalOpenCursor;

   function DataType2SqlType( aParam: TmParam): SQLSMALLINT;
   begin
     result := 0;
     case aParam.DataType of
             ftString:   result := SQL_CHAR;
             ftWord:     result := SQL_SMALLINT;
             ftSmallint: result := SQL_SMALLINT;
             ftInteger,
             ftAutoInc:  result := SQL_INTEGER;
//             ftTime:
             ftDate:     result := SQL_TYPE_DATE;
             ftDateTime: result := SQL_TYPE_TIMESTAMP;
             ftBCD,
             ftCurrency,
             ftFloat:    result := SQL_DOUBLE;
             ftBoolean:  result := SQL_SMALLINT;
             ftMemo:     result := SQL_CHAR;
//             ftUnknown:  DatabaseErrorFmt( SmFieldUndefinedType, [Name]);
             else        DatabaseErrorFmt( SmFieldUnsupportedType, [Name]);
     end;
   end;
  Var
     buf,
     bufPos:  Pchar;
     bufLen:  LongInt;
     dtype:   SQLSMALLINT;
     dsize:   SQLSMALLINT;
     i:       Integer;
     snd:     SQLINTEGER;
begin
   FieldDefs.Clear;
   FRecordSize:=0;
   InitStmt;
   try
     PrepareCursor;
     // seting MasterSource parameters
     if FDataLink.DataSource <> nil then SetParamsFromCursor;

     // calculate buffer length
     bufLen:=0;
     for i:=0 to Params.Count-1 do
       if not Params[i].isnull then
        case Params[i].DataType of
        ftDate:     bufLen:=bufLen+Sizeof(SQL_DATE_STRUCT);
        ftDateTime: bufLen:=bufLen+Sizeof(SQL_TIMESTAMP_STRUCT);
        else
           bufLen:=bufLen+Params[i].GetDataSize;
        end;

     if bufLen>0
       then GetMem( buf, bufLen)// buf:=StrAlloc(bufLen)
       else buf:=nil;

     // bind parameters
     try
        bufPos:=buf;
        for i:=0 to Params.Count-1 do
           begin
              dtype := DataType2SqlType( Params[i]);
              if (not Params[i].IsNull) then
               begin
                  case dtype of
                  SQL_TYPE_DATE:
                  begin
                       dsize := Sizeof(SQL_DATE_STRUCT);
                       PSQL_DATE_STRUCT( BufPos)^ :=
                                    DateTimeToDateStruct( Params[i].AsDateTime);
                  end;
                  SQL_TYPE_TIMESTAMP:
                  begin
                       dsize := Sizeof(SQL_TIMESTAMP_STRUCT);
                       DateTime2timeStampStruct( PSQL_TIMESTAMP_STRUCT( bufPos)^,
                                                 Params[i].AsDateTime);
                  end;
                  else
                       dsize:=Params[i].GetDataSize;
                       Params[i].GetData( bufPos);
                  end;
                  CheckSQLResult( SQLBindParameter( hstmt, i+1, SQL_PARAM_INPUT, SQL_C_DEFAULT{dtype}, dtype, dsize, 0, bufPos, 0, nil));
                  bufPos := bufPos + dsize;
               end else begin
                  snd:=SQL_NULL_DATA;
                  CheckSQLResult( SQLBindParameter( hstmt, i+1, SQL_PARAM_INPUT, SQL_C_DEFAULT{dtype}, dtype, 0, 0, nil, 0, @snd));
               end;
           end;
        try
           CheckSQLResult( SQLExecute( hstmt));
        except
           on E: ESQLerror do
                // change cursor type according stmt attributes
                if E.SqlState='01S02' then GetStmtAttributes else raise;
        end;
        Include(FState, mqOpened);
     finally
        if Params.Count>0 then
              CheckSQLResult( SQLFreeStmt( hstmt, SQL_RESET_PARAMS));
        if bufLen>0 then
              FreeMem( buf);
     end;
   except
     FreeStmt;
     raise;
   end;
end;

function TmCustomQuery.IsCursorOpen: Boolean;
begin
    Result:= (hstmt <> 0) and (mqOpened in FState);
end;

function TmCustomQuery.AllocRecordBuffer: PChar;
begin
  Result := StrAlloc(FRecBufSize);
  if FBlobCount > 0 then
    Initialize(PmBlobDataArray(Result + FBlobCacheOfs)[0], FBlobCount);
end;

procedure TmCustomQuery.FreeRecordBuffer(var Buffer: PChar);
begin
    if FBlobCount > 0 then
          Finalize(PmBlobDataArray(Buffer + FBlobCacheOfs)[0], FBlobCount);
    StrDispose(Buffer);
end;

procedure TmCustomQuery.SQLBindColumns( Buffer: PChar);
Var
   i: Integer;
begin
    if FBindBuffer = nil then FBindBuffer := AllocRecordBuffer;
    for I := 0 to FieldDefs.Count - 1 do
    with TODBCFieldDef(FieldDefs[I]) do
       if SQLDataType<>SQL_BINARY then
       CheckSQLResult( SQLBindCol( hstmt, FieldNo, SQLDataType,
                                   SQLPOINTER(FBindBuffer+OffsetInBuf),SqlSize,
                                   PSQLINTEGER(FBindBuffer+OffsetInBuf+SqlSize)));
end;

procedure TmCustomQuery.SQLUnBindColumns;
Var
   i: Integer;
   sqlres: SQLRETURN;
begin
   if FBindBuffer <> nil then
    begin
       sqlres:=SQLFreeStmt( hstmt, SQL_UNBIND);
       if (sqlres<>SQL_SUCCESS)and(sqlres<>SQL_INVALID_HANDLE) then
       for I := 0 to FieldDefs.Count - 1 do
          with TODBCFieldDef(FieldDefs[I]) do
          if SQLDataType<>SQL_BINARY then
              {CheckSQLResult(} SQLBindCol( hstmt, FieldNo, SqlDataType,
                                          nil, SQLSize, nil){)};
       FreeRecordBuffer( FBindBuffer);
       FBindBuffer := nil;
    end;
end;

function TmCustomQuery.GetRecord(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
var
   sqlres: SQLRETURN;
   i: integer;
   bind: SQLINTEGER;
   c:   AnsiString;
   iv:  SQLINTEGER;
   b:   LongInt;
label TryAgain;
begin
TryAgain:
  if FBindBuffer=nil then SQLBindColumns( FBindBuffer);
  case GetMode of
    gmCurrent:
      case FetchStatus of
      bfBOF: sqlres := SQLFetchScroll( hstmt, SQL_FETCH_FIRST, 0);
      bfEOF: sqlres := SQLFetchScroll( hstmt, SQL_FETCH_LAST, 0);
      else   sqlres := SQLFetchScroll( hstmt, SQL_FETCH_RELATIVE, 0);
      end;
//      sqlres := SQLSetPos( hstmt, 1, SQL_POSITION, SQL_LOCK_NO_CHANGE);
    gmNext:
      case FetchStatus of
      bfBOF: sqlres := SQLFetchScroll( hstmt, SQL_FETCH_FIRST, 0);
      bfEOF: sqlres := SQL_NO_DATA; //SQL_ERROR;
      else   sqlres := SQLFetchScroll( hstmt, SQL_FETCH_NEXT, 1);
      end;
//      sqlres := SQLExtendedFetch( hstmt, SQL_FETCH_NEXT, 1, @rCount, @rStatus);
    gmPrior:
      case FetchStatus of
      bfBOF: sqlres := SQL_NO_DATA; //SQL_ERROR;
      bfEOF: sqlres := SQLFetchScroll( hstmt, SQL_FETCH_LAST, 0);
      else   sqlres := SQLFetchScroll( hstmt, SQL_FETCH_PRIOR, 1);
      end;
  else
    sqlres:= SQL_ERROR;
  end;
  FetchStatus:=bfCurrent;

  Result := grOK;
  case sqlres of
     SQL_SUCCESS:;
     SQL_SUCCESS_WITH_INFO:
       try
          checkSQLResult(sqlres);
       except on E: ESQLerror do
               if (E.SqlState<>'01S06')and(GetMode=gmPrior)
                 then Result:=grBOF
                 else raise;
       end;
     SQL_NEED_DATA,
     SQL_STILL_EXECUTING,
     SQL_ERROR,
     SQL_INVALID_HANDLE: begin checkSQLResult(sqlres); Result:=grError; end;
     SQL_NO_DATA:
           if GetMode=gmNext then Result:=grEOF
           else if GetMode=gmPrior then Result:=grBOF
                else Result:=grError;{raise ESQLerror.Create('SQL_NO_DATA');}
     else                   Result:=grError;
     end;
  if Result<>grOk then exit;
  if CursorAttr.UseDriverBookMark // read bookmark
    then CheckSQLResult( SQLGetData( hstmt, 0, SQL_BINARY,
                                  SQLPOINTER(Buffer),BookMarkSize,@bind));
  with PODBCRecInfo(Buffer + FRecInfoOfs)^ do
    begin
       BookmarkFlag := bfCurrent;
       if SQLGetStmtAttr( hstmt, SQL_ATTR_ROW_NUMBER, @RecordNumber, SQL_IS_UINTEGER, nil)
             <>SQL_SUCCESS then RecordNumber:=-1;
       if RecordNumber=0 then RecordNumber:=-1;
//       UpdateStatus := TUpdateStatus(FRecProps.iRecStatus);
    end;

  // check deleted and filtered records
  if not CheckDeleted(Buffer) then
  begin
       if GetMode=gmCurrent then GetMode:=gmNext;
       goto TryAgain;
  end;

  try
    // check if record cached
    b:=-1;
    for i:=0 to FCachedCount-1 do
      if MemCompare( getBookmarkPtr( Buffer),
                     getBookmarkPtr( FCachedBuffer[i]), BookMarkSize)
        then begin b:=i; break; end;
    if b >= 0 then
      begin // read fields into buffer from cache
         CopyRecordBuffer( FCachedBuffer[b], Buffer);
      end
    else begin// read fields into buffer from driver
       if CursorAttr.UseDriverBookMark
            then Move( (FBindBuffer+BookMarkSize)^,
                       (Buffer+BookMarkSize)^, FRecordSize-BookMarkSize) // copy only binded fields without blob
            else Move( FBindBuffer^, Buffer^, FRecordSize); // copy only binded fields without blob
       for I := 0 to FieldDefs.Count - 1 do
       with TODBCFieldDef(FieldDefs[I]) do
       if SQLDataType=SQL_BINARY then {
       CheckSQLResult( SQLGetData( hstmt,FieldNo, SQLDataType,
                                   SQLPOINTER(Buffer+OffsetInBuf),SqlSize,
                                   PSQLINTEGER(Buffer+OffsetInBuf+SqlSize)))
       else }
       begin
          // load blob fields
          c:='';
          sqlres:=SQLGetData( hstmt, FieldNo, SQL_BINARY, 
                              SQLPOINTER( @i), 0, @iv); // get data length
          if (sqlres=SQL_SUCCESS)
             or(sqlres=SQL_SUCCESS_WITH_INFO) then
          begin
            SetLength( c, iv+1); // one more byte for #0
            sqlres:=SQLGetData( hstmt, FieldNo, SQL_BINARY, 
                                SQLPOINTER( PCHAR(c)), iv+1, @iv);
            if (sqlres <> SQL_SUCCESS)and(sqlres <> SQL_NO_DATA)
             then CheckSQLResult( sqlres);
            SetLength( c, iv); // set length to real getted data
          end;
          {$RANGECHECKS OFF}
          PmBlobDataArray(Buffer + FBlobCacheOfs)[OffsetInBuf]:=c;
       end;
    end;
    GetCalcFields(Buffer);
  except
     Result:=grError;
  end;
  if not CheckFilter(Buffer) then
  begin
       if GetMode=gmCurrent then GetMode:=gmNext;
       goto TryAgain;
  end;
end;

procedure TmCustomQuery.InternalSetToRecord(Buffer: PChar);
begin
  InternalGotoBookmark( getBookmarkPtr( Buffer));
end;

function TmCustomQuery.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  RecBuf: PChar;
  fd:     TODBCFieldDef;
  b:      ^TDateTimeRec;
  i:      longint;
begin
  Result := False;
  if not GetActiveRecBuf(RecBuf) then Exit;
  with Field do
    if FieldNo > 0 then
    begin
       fd:=TODBCFieldDef(FieldDefs.Find(FieldName));
       if PSQLINTEGER(RecBuf+fd.OffsetInBuf+fd.SqlSize)^=SQL_NULL_DATA then Exit;
       if Buffer<>nil then
       case DataType of
       ftString:
          begin
               Move( (RecBuf+fd.OffsetInbuf)^, Buffer^,fd.SqlSize);
               i:=StrLen( Buffer)-1;
               while (i>=0) and (PChar(Buffer)[i]=#32) do begin PChar(Buffer)[i]:=#0; dec(i); end;
          end;
       ftSmallint,
       ftInteger,
       ftFloat: Move( (RecBuf+fd.OffsetInbuf)^, Buffer^,fd.SqlSize);
       ftBoolean: PWord(Buffer)^ := PByte(RecBuf+fd.OffsetInbuf)^;
       ftDate:
          begin
               b:=Buffer;

               b.Date := DateTimeToTimeStamp(
                           DateStructToDateTime(
                             PSQL_DATE_STRUCT( RecBuf+fd.OffsetInbuf))).Date;
          end;
       ftTime:
          begin
               b:=Buffer;
               with PSQL_TIME_STRUCT( RecBuf+fd.OffsetInbuf)^ do
                   b.Time := DateTimeToTimeStamp(EncodeTime( Hour, Minute, Second, 0)).Time;
          end;
       ftDateTime:
               with PSQL_TIMESTAMP_STRUCT( RecBuf+fd.OffsetInbuf)^ do
               begin
                  b:=Buffer;
                  b^.DateTime:=TimeStampToMSecs(
                                 DateTimeToTimeStamp(
                                     EncodeDate( Year, Month, Day)+
                                     EncodeTime( Hour, Minute, Second, 0)));
               end;
       else
           raise Exception.CreateFmt( SmFieldUnsupportedType,[ FieldName]);
       end;
       Result:=True;
    end
    else
      if State in [dsBrowse, dsEdit, dsInsert, dsCalcFields] then
      begin
        Inc(RecBuf, FRecordSize + Offset);
        Result := Boolean(RecBuf[0]);
        if Result and (Buffer <> nil) then
          Move(RecBuf[1], Buffer^, DataSize);
      end;
end;

procedure TmCustomQuery.SetFieldData(Field: TField; Buffer: Pointer);
var
  RecBuf:     PChar;
  TimeStamp:  TTimeStamp;
  fd:         TODBCFieldDef;
  b:          ^TDateTimeRec;
begin
  with Field do
  begin
    if not (State in dsWriteModes) then DatabaseError(SNotEditing);
    GetActiveRecBuf(RecBuf);
    if FieldNo > 0 then
    begin
      if State = dsCalcFields then DatabaseError(SNotEditing);
      if ReadOnly and not (State in [dsSetKey, dsFilter]) then
        DatabaseErrorFmt(SFieldReadOnly, [DisplayName]);
      Validate(Buffer);
      if FieldKind <> fkInternalCalc then
      begin
//        if FConstraintLayer and Field.HasConstraints and (State in [dsEdit, dsInsert]) then
//          Check(DbiVerifyField(FHandle, FieldNo, Buffer, Blank));
//        Check(DbiPutField(FHandle, FieldNo, RecBuf, Buffer));
       fd:=TODBCFieldDef(FieldDefs.Find(FieldName));
       if Buffer=nil then
       begin // field set to NULL
            PSQLINTEGER(RecBuf+fd.OffsetInBuf+fd.SqlSize)^:=SQL_NULL_DATA;
       end else
       case fd.SQLDataType of
       SQL_CHAR:
       begin
            StrCopy(RecBuf+fd.OffsetInBuf, Buffer);
            PSQLINTEGER(RecBuf+fd.OffsetInBuf+fd.SqlSize)^:=StrLen(Buffer);
       end;
       SQL_SMALLINT:
       begin
            Move( Buffer^, (RecBuf+fd.OffsetInBuf)^,sizeof(SQLSMALLINT));
            PSQLINTEGER(RecBuf+fd.OffsetInBuf+fd.SqlSize)^:=sizeof(SQLSMALLINT);
       end;
       SQL_INTEGER:
       begin
            Move( Buffer^, (RecBuf+fd.OffsetInBuf)^,sizeof(SQLINTEGER));
            PSQLINTEGER(RecBuf+fd.OffsetInBuf+fd.SqlSize)^:=sizeof(SQLINTEGER);
       end;
       SQL_BIT:
       begin
            PByte(RecBuf+fd.OffsetInbuf)^ := PWord(Buffer)^;
            PSQLINTEGER(RecBuf+fd.OffsetInBuf+fd.SqlSize)^:=sizeof(SQLCHAR);
       end;
       SQL_DOUBLE:
       begin
            Move( Buffer^, (RecBuf+fd.OffsetInBuf)^,sizeof(SQLDOUBLE));
            PSQLINTEGER(RecBuf+fd.OffsetInBuf+fd.SqlSize)^:=sizeof(SQLINTEGER);
       end;
       SQL_TYPE_DATE:
       begin
            b:=Buffer;
            TimeStamp.Date:=b^.Date;
            PSQL_DATE_STRUCT( RecBuf+fd.OffsetInbuf)^:=
               DateTimeToDateStruct( TimeStampToDateTime(TimeStamp));
            PSQLINTEGER(RecBuf+fd.OffsetInBuf+fd.SqlSize)^:=sizeof(SQL_DATE_STRUCT);
       end;
       SQL_TYPE_TIMESTAMP:
          begin
               b:=Buffer;
               DateTime2timeStampStruct( PSQL_TIMESTAMP_STRUCT( RecBuf+fd.OffsetInbuf)^,
                                         TimeStampToDateTime(MSecsToTimeStamp(b.DateTime)));
               PSQLINTEGER(RecBuf+fd.OffsetInBuf+fd.SqlSize)^:=sizeof(SQL_TIMESTAMP_STRUCT);
          end;

       else raise Exception.CreateFmt( SmFieldUnupdatedType,[ FieldName]);
       end;
      end;
    end else {fkCalculated, fkLookup}
    begin
      Inc(RecBuf, FRecordSize + Offset);
      Boolean(RecBuf[0]) := LongBool(Buffer);
      if Boolean(RecBuf[0]) then Move(Buffer^, RecBuf[1], DataSize);
    end;
    if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
      DataEvent(deFieldChange, Longint(Field));
  end;
end;

function TmCustomQuery.GetActiveRecBuf(var RecBuf: PChar): Boolean;
begin
  case State of
    dsBrowse: if IsEmpty then RecBuf := nil else RecBuf := ActiveBuffer;
    dsEdit, dsInsert: RecBuf := ActiveBuffer;
//    dsSetKey: RecBuf := PChar(FKeyBuffer) + SizeOf(TKeyBuffer);
    dsCalcFields: RecBuf := CalcBuffer;
    dsFilter: RecBuf := FFilterBuffer;
    dsNewValue: RecBuf := ActiveBuffer;
    dsOldValue: if fOldBuffer<>nil
                  then RecBuf := FOldBuffer
                  else RecBuf := ActiveBuffer;
  else
    RecBuf := nil;
  end;
  Result := RecBuf <> nil;
end;

procedure TmCustomQuery.InternalFirst;
//var
//   sqlres: SQLRETURN;
begin
    FetchStatus := bfBOF;
{//   SQLUnBindColumns;
   Screen.Cursor:=crSQLWait;
   try
     sqlres := SQLFetchScroll( hstmt, SQL_FETCH_FIRST, 0);
     case sqlres of
       SQL_SUCCESS:;
       SQL_SUCCESS_WITH_INFO: raise ESQLerror.CreateDiag(SQL_HANDLE_STMT,hstmt, sqlres);
       SQL_NEED_DATA:         raise ESQLerror.Create('SQL_NEED_DATA');
       SQL_STILL_EXECUTING:   raise ESQLerror.Create('SQL_STILL_EXECUTING');
       SQL_ERROR:             raise ESQLerror.CreateDiag(SQL_HANDLE_STMT,hstmt, sqlres);
       SQL_NO_DATA:           ;//raise ESQLerror.Create('SQL_NO_DATA');
       SQL_INVALID_HANDLE:    raise ESQLerror.Create('SQL_INVALID_HANDLE');
     else                   raise ESQLerror.Create('unknown SQL result');
     end;
     SQLFetchScroll( hstmt, SQL_FETCH_PRIOR, 0);
   finally
      Screen.Cursor:=crDefault;
   end; }
end;

procedure TmCustomQuery.InternalLast;
begin
     FetchStatus:=bfEOF;
//   SQLUnBindColumns;
{   Screen.Cursor:=crSQLWait;
   try
      CheckSQLResult( SQLFetchScroll( hstmt, SQL_FETCH_LAST, 0));
      SQLFetchScroll( hstmt, SQL_FETCH_NEXT, 0);
   finally
      Screen.Cursor:=crDefault;
   end; }
end;

procedure TmCustomQuery.InternalInitRecord(Buffer: PChar);
var
   i:integer;
begin
     for I := 0 to BlobFieldCount - 1 do
       PmBlobDataArray(Buffer + FBlobCacheOfs)[I] := '';
     // clear indicator var //FillChar( Buffer^, FRecordSize, #0);
     for I := 0 to FieldDefs.Count - 1 do
       with TODBCFieldDef(FieldDefs[I]) do
          PSQLINTEGER(Buffer+OffsetInBuf+SqlSize)^:=SQL_NULL_DATA;
     with PODBCRecInfo(Buffer + FRecInfoOfs)^ do
     begin
          UpdateStatus := TUpdateStatus(usInserted);
          BookMarkFlag := bfInserted;
          RecordNumber:=-1;
     end;
end;

procedure TmCustomQuery.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
end;

procedure TmCustomQuery.InternalEdit;
begin
     if FoldBuffer<>nil then FreeRecordBuffer(FoldBuffer);
     FoldBuffer := AllocRecordBuffer;
     CopyRecordBuffer( ActiveBuffer, FoldBuffer);
end;

function TmCustomQuery.GetUpdateQuery(UpdateKind: TUpdateKind): TmCustomQuery;
  begin
    if not Assigned(FQueries[UpdateKind]) then
    begin
      FQueries[UpdateKind] := TmCustomQuery.Create(Self);
      FQueries[UpdateKind].Database := Self.DataBase;
      FQueries[UpdateKind].CursorType := Self.CursorType;
    end;
    Result := FQueries[UpdateKind];
  end;

procedure TmCustomQuery.SetParamsForUpdateSql(UpdateKind: TUpdateKind);
var
  I: Integer;
  Old: Boolean;
  Param: TmParam;
  PName: string;
  Field: TField;
  Value: Variant;
begin
  if Self.FUpdateSQL[UpdateKind].Count=0 then raise Exception.Create( SmStatementUndefined);
  with GetUpdateQuery( UpdateKind) do
  begin
    SQL.Assign(Self.FUpdateSQL[UpdateKind]);
    for I := 0 to Params.Count - 1 do
    begin
      Param := Params[I];
      PName := Param.Name;
      Old := CompareText(Copy(PName, 1, 4), 'OLD_') = 0;
      if Old then System.Delete(PName, 1, 4);
      Field := Self.FindField(PName);
      if not Assigned(Field) then Continue;
      if Old then Param.AssignFieldValue(Field, Field.OldValue) else
      begin
        Value := Field.NewValue;
        if VarIsEmpty(Value) then Value := Field.OldValue;
        Param.AssignFieldValue(Field, Value);
      end;
    end;
  end;
end;

procedure TmCustomQuery.InternalPost;
var
   i:     integer;
   iNull: SQLINTEGER;
   bNull: char;
  procedure BindCurRecord;
  var
   i:        integer;
   FieldDef: TODBCFieldDef;
  begin
     SQLUnBindColumns;
     if CursorAttr.UseDriverBookMark
          then CheckSQLResult( SQLBindCol( hstmt, 0, SQL_BINARY,
                                           SQLPOINTER(ActiveBuffer),
                                           BookMarkSize,nil));
     for I := 0 to FieldCount - 1 do
       with TField(Fields[I]) do begin
          if (FieldKind<>fkData)
             or(ReadOnly=True)
             or(FieldNo<=0) then continue;
          FieldDef := FieldDefs.Find(FieldName) as TODBCFieldDef;
          if FieldDef=nil then raise Exception.create('fielddef not found');
          with FieldDef do
          if SQLDataType<>SQL_BINARY then
             CheckSQLResult( SQLBindCol( hstmt, FieldNo, SQLDataType,
                             SQLPOINTER(ActiveBuffer+OffsetInBuf),SqlSize,
                             PSQLINTEGER(ActiveBuffer+OffsetInBuf+SqlSize)))
          else
             if Length(PmBlobDataArray(ActiveBuffer + FBlobCacheOfs)[OffsetInBuf])>0
              then CheckSQLResult( SQLBindCol( hstmt, FieldNo, SQLDataType,
                        SQLPOINTER(PmBlobDataArray(ActiveBuffer + FBlobCacheOfs)[OffsetInBuf]),
                        Length(PmBlobDataArray(ActiveBuffer + FBlobCacheOfs)[OffsetInBuf]),
                        nil))
              else begin
                CheckSQLResult( SQLBindCol( hstmt, FieldNo, SQLDataType,
                        SQLPOINTER(@bNull),
                        0,
                        SQLPOINTER(@iNull)));
              end;
       end;
{     for I := 0 to FieldDefs.Count - 1 do
        with TODBCFieldDef(FieldDefs[I]) do
        if SQLDataType<>SQL_BINARY then
          CheckSQLResult( SQLBindCol( hstmt, FieldNo, SQLDataType,
                        SQLPOINTER(ActiveBuffer+OffsetInBuf),SqlSize,
                        PSQLINTEGER(ActiveBuffer+OffsetInBuf+SqlSize)))
        else
          if Length(PmBlobDataArray(ActiveBuffer + FBlobCacheOfs)[OffsetInBuf])>0
           then CheckSQLResult( SQLBindCol( hstmt, FieldNo, SQLDataType,
                        SQLPOINTER(PmBlobDataArray(ActiveBuffer + FBlobCacheOfs)[OffsetInBuf]),
                        Length(PmBlobDataArray(ActiveBuffer + FBlobCacheOfs)[OffsetInBuf]),
                        nil))
           else begin
                CheckSQLResult( SQLBindCol( hstmt, FieldNo, SQLDataType,
                        SQLPOINTER(@bNull),
                        0,
                        SQLPOINTER(@iNull)));
           end;
}
  end;
  procedure UnBindCurRecord;
  Var
   i: Integer;
   sqlres: SQLRETURN;
  begin
     if CursorAttr.UseDriverBookMark
         then SQLBindCol( hstmt, 0, SQL_BINARY,nil,0,PSQLINTEGER(0));
     sqlres:=SQLFreeStmt( hstmt, SQL_UNBIND);
     if (sqlres<>SQL_SUCCESS)and(sqlres<>SQL_INVALID_HANDLE) then
     for I := 0 to FieldDefs.Count - 1 do
        with TODBCFieldDef(FieldDefs[I]) do
//        if SQLDataType<>SQL_BINARY then
            CheckSQLResult( SQLBindCol( hstmt, FieldNo, SqlDataType,
                                        nil, SQLSize, nil));
  end;
begin
   iNull:=SQL_NULL_DATA;
   bNull:=#0;
   if CachedUpdates then
   begin
        AddToCacheBuffer(usModified, ActiveBuffer);
        Exit;
   end;

      case State of
        dsInsert: begin
           if Length(FUpdateSQL[ukInsert].Text)>0 then begin
                 // use updatesql
                 SetParamsForUpdateSQL(ukInsert);
                 FQueries[ukInsert].ExecSql;
                 i:=FQueries[ukInsert].RowsAffected;
                 if i<=0 then ShowMessage( SmNoRowsAffected);
                 if i>1  then ShowMessage( SmMoreRowsAffected);
              end else
           if (FCursorAttr.Attrib1 and SQL_CA1_BULK_ADD)<>0 then begin
                 BindCurRecord;
                 try
                    CheckSQLResult( SQLBulkOperations( hstmt, SQL_ADD));
                    // for ODBC 3.0 drivers bookmark setted for new record ... but not work ;)
                 finally
                    UnBindCurRecord;
                 end;
             end
             else raise Exception.Create( SmInsertNotSupported);
//           if (FCursorAttr.Sensitivity and SQL_SS_ADDITIONS)=0
//             then AddToCacheBuffer( usInserted, ActiveBuffer); in the future
        end;
      dsEdit: begin
           if Length(FUpdateSQL[ukModify].Text)>0 then begin
                 // use updatesql
                 SetParamsForUpdateSQL(ukModify);
                 FQueries[ukModify].ExecSql;
                 i:=FQueries[ukModify].RowsAffected;
                 if i<=0 then ShowMessage( SmNoRowsAffected);
                 if i>1  then ShowMessage( SmMoreRowsAffected);
              end else
           if (FCursorAttr.Attrib1 and SQL_CA1_POS_UPDATE)<>0 then begin
                 BindCurRecord;
                 try
                    CheckSQLResult( SQLSetPos( hstmt, 1, SQL_UPDATE, SQL_LOCK_NO_CHANGE ));
                    //CheckSQLResult( SQLBulkOperations( hstmt, SQL_UPDATE_BY_BOOKMARK));
                 finally
                    UnBindCurRecord;
                 end;
             end
             else raise Exception.Create( SmUpdateNotSupported);
           if FoldBuffer<>nil then FreeRecordBuffer( FoldBuffer);
           FoldBuffer := nil;
           if (FCursorAttr.Sensitivity and SQL_SS_UPDATES)=0
             then AddToCacheBuffer( usModified, ActiveBuffer);
        end;
      end;
end;

procedure TmCustomQuery.InternalCancel;
begin

end;

procedure TmCustomQuery.InternalDelete;
Var
   i:  LongInt;
begin
   if Length(FUpdateSQL[ukDelete].Text)>0 then begin
         // use updatesql
         SetParamsForUpdateSQL(ukDelete);
         FQueries[ukDelete].ExecSql;
         i:=FQueries[ukDelete].RowsAffected;
         if i<=0 then ShowMessage( SmNoRowsAffected);
         if i>1  then ShowMessage( SmMoreRowsAffected);
     end else
   if (FCursorAttr.Attrib1 and SQL_CA1_POS_DELETE)<>0 then
         CheckSQLResult( SQLSetPos( hstmt, 1, SQL_DELETE, SQL_LOCK_NO_CHANGE))
     else raise Exception.Create( SmDeleteNotSupported);
   if (FCursorAttr.Sensitivity and SQL_SS_DELETIONS)=0
     then AddDeletedToList( ActiveBuffer);
end;

{ BookMarks }
procedure TmCustomQuery.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
   Move(getBookmarkPtr( Buffer)^, Data^, BookmarkSize);
end;

procedure TmCustomQuery.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
{   if CursorAttr.UseDriverBookMark
   then Move( Data^, getBookmarkPtr( Buffer)^, BookmarkSize)
   else} Move( Data^, getBookmarkPtr( Buffer)^, BookmarkSize);
end;

function TmCustomQuery.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := PODBCRecInfo(Buffer + FRecInfoOfs).BookmarkFlag;
end;

procedure TmCustomQuery.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  PODBCRecInfo(Buffer + FRecInfoOfs).BookmarkFlag := Value;
end;

procedure TmCustomQuery.InternalGotoBookmark(Bookmark: Pointer);
Var
   i: SQLINTEGER;
begin
//   SQLUnBindColumns;
   if CursorAttr.UseDriverBookMark then
   begin
      CheckSQLResult( SQLSetStmtAttr( hstmt, SQL_ATTR_FETCH_BOOKMARK_PTR, SQLPOINTER(BookMark), SQL_IS_POINTER));
      CheckSQLResult( SQLFetchScroll( hstmt, SQL_FETCH_BOOKMARK, 0));
   end else
   begin
      if (CursorAttr.Attrib1 and SQL_CA1_ABSOLUTE)>0 then begin
                  i:=PSQLINTEGER(BookMark)^;
                  if i<0 then raise Exception.Create( SmUndefinedRecNo);
                  CheckSQLResult( SQLFetchScroll( hstmt, SQL_FETCH_ABSOLUTE, i));
      end else raise Exception.Create( SmErrAbsSetToRec);
//      i:=PSQLINTEGER(BookMark)^;
//      CheckSQLResult( SQLFetchScroll( hstmt, SQL_FETCH_ABSOLUTE, i));
   end;
end;

function TmCustomQuery.GetCanModify: Boolean;
begin
  Result := FCanModify;
end;

procedure TmCustomQuery.QueryChanged(Sender: TObject);
var
  List: TmParams;
begin
  FText := SQL.Text;
  if not (csLoading in ComponentState) then
  begin
    Close;
    UnPrepareCursor;
//    StrDispose(SQLBinary);
//    SQLBinary := nil;
    if ParamCheck or (csDesigning in ComponentState) then
    begin
      List := TmParams.Create;
      try
        CreateParams(List, PChar(Text));
        List.AssignValues(FParams);
        FParams.Free;
        FParams := List;
      except
        List.Free;
      end;
    end;
    DataEvent(dePropertyChange, 0);
  end else
    CreateParams(nil, PChar(Text));
end;

procedure TmCustomQuery.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FDataBase) then
    FDataBase := nil;
{  if (Operation = opRemove) and (AComponent = FUpdateObject) then
    FUpdateObject := nil; }
end;

function TmCustomQuery.GetCursorName:String;
Var
   curName: array [0..30] of Char;
   nl:      SQLSMALLINT;
begin
     CheckSQLResult( SQLGetCursorName( hstmt, curName, 30, @nl));
     Result:=StrPas( curNAme);
end;
{
function TmCustomQuery.GetRowsAffected: LongInt;
Var
   rCount: SQLINTEGER;
begin
     CheckSQLResult( SQLRowCount( hstmt, rCount));
     Result:= rCount;
end;
}

// ᮧ ᯨ᮪ ࠬ஢  SQL ப
//  塞 ࠬ  ᨬ "?"
procedure TmCustomQuery.CreateParams(List: TmParams; const Value: PChar);
var
  CurPos, StartPos: PChar;
  CurChar: Char;
  Literal: Boolean;
  EmbeddedLiteral: Boolean;
  Name: string;

  function NameDelimiter: Boolean;
  begin
    Result := CurChar in [' ', ',', ';', ')', #13, #10];
  end;

  function IsLiteral: Boolean;
  begin
    Result := CurChar in ['''', '"'];
  end;

  function StripLiterals(Buffer: PChar): string;
  var
    Len: Word;
    TempBuf: PChar;

    procedure StripChar(Value: Char);
    begin
      if TempBuf^ = Value then
        StrMove(TempBuf, TempBuf + 1, Len - 1);
      if TempBuf[StrLen(TempBuf) - 1] = Value then
        TempBuf[StrLen(TempBuf) - 1] := #0;
    end;

  begin
    Len := StrLen(Buffer) + 1;
    TempBuf := AllocMem(Len);
    Result := '';
    try
      StrCopy(TempBuf, Buffer);
      StripChar('''');
      StripChar('"');
      Result := StrPas(TempBuf);
    finally
      FreeMem(TempBuf, Len);
    end;
  end;

begin
  CurPos := Value;
  Literal := False;
  EmbeddedLiteral := False;
  repeat
    CurChar := CurPos^;
    if (CurChar = ':') and not Literal and ((CurPos + 1)^ <> ':') then
    begin
      StartPos := CurPos;
      while (CurChar <> #0) and (Literal or not NameDelimiter) do
      begin
        Inc(CurPos);
        CurChar := CurPos^;
        if IsLiteral then
        begin
          Literal := Literal xor True;
          if CurPos = StartPos + 1 then EmbeddedLiteral := True;
        end;
      end;
      CurPos^ := #0;
      if EmbeddedLiteral then
      begin
        Name := StripLiterals(StartPos + 1);
        EmbeddedLiteral := False;
      end
      else Name := StrPas(StartPos + 1);
      if Assigned(List) then
        List.CreateParam(ftUnknown, Name, ptUnknown);
      CurPos^ := CurChar;
      StartPos^ := '?';
      Inc(StartPos);
      StrMove(StartPos, CurPos, StrLen(CurPos) + 1);
      CurPos := StartPos;
    end
    else if (CurChar = ':') and not Literal and ((CurPos + 1)^ = ':') then
      StrMove(CurPos, CurPos + 1, StrLen(CurPos) + 1)
    else if IsLiteral then Literal := Literal xor True;
    Inc(CurPos);
  until CurChar = #0;
end;

procedure TmCustomQuery.SetParamsList(Value: TmParams);
begin
  FParams.AssignValues(Value);
end;

function TmCustomQuery.GetRecordCount: Integer;
// this function could be sloooow,
// try to _not_ use it
var
   BM: TBookmark;
   count: integer;
begin
     BM := GetBookmark;
     DisableControls;
     count := 0;
     First;
     while not EOF do
     begin
          count := count + 1;
          Next;
     end;
     GotoBookmark( BM);
     FreeBookMark( BM);
     EnableControls;
     Result := count;
//     Result:=-1;
end;

function TmCustomQuery.GetRecNo: Integer;
var
  BufPtr: PChar;
begin
  CheckActive;
  if State = dsCalcFields then
    BufPtr := CalcBuffer else
    BufPtr := ActiveBuffer;
  Result := PODBCRecInfo(BufPtr + FRecInfoOfs).RecordNumber;
end;

function TmCustomQuery.IsSequenced: Boolean;
begin
  Result := False;
end;

procedure TmCustomQuery.RefreshParams;
var
  DataSet: TDataSet;
begin
  DisableControls;
  try
    if (FDataLink.DataSource <> nil)
       and(not (csDesigning in ComponentState)) then
    begin
      DataSet := FDataLink.DataSource.DataSet;
      if DataSet <> nil then
        if DataSet.Active and (DataSet.State <> dsSetKey) then
        begin
          Close;
          Open;
        end;
    end;
  finally
    EnableControls;
  end;
end;

procedure TmCustomQuery.SetDataSource(Value: TDataSource);
begin
  if IsLinkedTo(Value) then DatabaseError(SCircularDataLink);
  FDataLink.DataSource := Value;
end;

function TmCustomQuery.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TmCustomQuery.SetParamsFromCursor;
var
  I: Integer;
  DataSet: TDataSet;
begin
  if FDataLink.DataSource <> nil then
  begin
    DataSet := FDataLink.DataSource.DataSet;
    if DataSet <> nil then
    begin
      DataSet.FieldDefs.Update;
      for I := 0 to FParams.Count - 1 do
        with FParams[I] do
          if not Bound then
          begin
            if DataSet.Active
              then AssignField(DataSet.FieldByName(Name))
              else DataType:=DataSet.FieldDefs.Find(Name).DataType;
            Bound := False;
          end;
    end;
  end;
end;

procedure TmCustomQuery.SetCachedUpdates(Value: Boolean);
begin
  if (State = dsInActive) or (csDesigning in ComponentState) then
    FCachedUpdates := Value
  else
    raise Exception.Create('Can''t change CachedUpdates property' );
end;

procedure TmCustomQuery.ApplyUpdates;
begin
end;

procedure TmCustomQuery.CommitUpdates;
begin
end;

procedure TmCustomQuery.CancelUpdates;
begin
end;

procedure TmCustomQuery.SetCursorType(Value: TmCursorTypes);
begin
     if Value<>FCursorType then
     begin
          CheckInactive;
          FCursorType:=Value;
     end;
end;

function TmCustomQuery.ParamByName(const Value: string): TmParam;
begin
  Result := FParams.ParamByName(Value);
end;

{ Blobs }

function TmCustomQuery.GetBlobData(Field: TField; Buffer: PChar): TmBlobData;
begin
  Result := PmBlobDataArray(Buffer + FBlobCacheOfs)[Field.Offset];
end;

procedure TmCustomQuery.SetBlobData(Field: TField; Buffer: PChar; Value: TmBlobData);
begin
  if Buffer = ActiveBuffer then
    PmBlobDataArray(Buffer + FBlobCacheOfs)[Field.Offset] := Value;
end;

function TmCustomQuery.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
begin
   Result := TmBlobStream.Create(Field as TBlobField, Mode);
end;

function TmCustomQuery.Locate(const KeyFields: string;
           const KeyValues:Variant; Options: TLocateOptions): Boolean;
var
  SearchFields: TList;
  I:Integer;
  Match:boolean;

Function MatchVariants( FieldType:TFieldType; m1,M2:Variant; 
			CaseSensitive:boolean; PartialKey:boolean):boolean;
var
   s1,s2:string;
begin
   if fieldType=ftString then begin
       S1:=M1;
       S2:=M2;
       if not CaseSensitive then
         begin
            S1 := AnsiUpperCase( S1);
            S2 := AnsiUpperCase( S2);
         end;
       if PartialKey 
	 then result := (pos( S2, S1) > 0)
         else result := (S1 = S2);  
   end else begin
       result:=(m1=m2);
   end;
end;
begin
   DoBeforeScroll;
   CheckBrowseMode;
   CursorPosChanged;

   Result := False;
   SearchFields := TList.Create;
   try
       GetFieldList(SearchFields, KeyFields);
       InternalFirst;
       while (GetRecord( ActiveBuffer, gmNext, True) = grOK) do begin
        if Searchfields.Count=1 then begin
            match:=MatchVariants(TField(SearchFields[0]).DataType,
                     FieldByName(TField(SearchFields[0]).FieldName).Value,KeyValues,
                     (loCaseInsensitive in Options),
                     (loPartialKey in Options));
        end else begin
            match:=true;
            for i:=0 to Searchfields.Count-1 do begin
                        match:=match and MatchVariants(TField(SearchFields[i]).DataType,
                                            FieldByName(TField(SearchFields[i]).FieldName).Value,KeyValues[i],
                                            (loCaseInsensitive in Options), 
                                            (loPartialKey in Options));
            end;
        end;
        if match then begin
               Resync([rmExact, rmCenter]);
//               DoAfterScroll;
               Result:=True;
               break;
        end;
       end;
   finally
       DoAfterScroll;
       SearchFields.Free;
   end;
end;

{
function TmCustomQuery.Locate(const KeyFields: string; const KeyValues: Variant;
  Options: TLocateOptions): Boolean;
var
  SaveState: TDataSetState;
begin
   DoBeforeScroll;
   CheckBrowseMode;
   CursorPosChanged;

   Result := False;
//   SaveState := SetTempState(dsFilter);
//   FFilterBuffer := AllocRecordBuffer;
   try
      InternalFirst;
      while (GetRecord( ActiveBuffer, gmNext, True) = grOK) do
        begin
            if FieldByName(KeyFields).Value=KeyValues then begin
//               RestoreState(SaveState);
//               FreeRecordBuffer( FFilterBuffer);
//               FFilterBuffer:=nil;
               Resync([rmExact, rmCenter]);
               DoAfterScroll;
               Result:=True;
               break;
            end;
        end;
   finally
//      Resync([rmExact, rmCenter]);
      DoAfterScroll;
//      RestoreState(SaveState);
//      FreeRecordBuffer( FFilterBuffer);
//      FFilterBuffer:=nil;
   end;
end;


function TmCustomQuery.Locate(const KeyFields: string; const KeyValues: Variant;
  Options: TLocateOptions): Boolean;
  Var
     SaveState: TDataSetState;
begin
  SaveState := SetTempState(dsFilter);
  InternalFirst;
  FFilterBuffer := AllocRecordBuffer;
  try
  while (GetRecord(FFilterBuffer, gmNext, True) = grOK) do
    begin
      // ஢ઠ KeyValues
      if FieldByName(KeyFields).Value = KeyValues then
        begin
           RestoreState(SaveState);
           FreeRecordBuffer( FFilterBuffer);
           Resync([rmCenter]);
           Result := True;
           exit;
        end;
    end;
  finally
    RestoreState(SaveState);
    FreeRecordBuffer( FFilterBuffer);
  end;
  Result := False;
end;
}

function TmCustomQuery.Lookup(const KeyFields: string; const KeyValues: Variant;
      const ResultFields: string): Variant;
begin
{  Result := Null;
  if Locate(KeyFields, KeyValues, []) then
  begin
    SetTempState(dsCalcFields);
    try
      CalculateFields(TempBuffer);
      Result := FieldValues[ResultFields];
    finally
      RestoreState(dsBrowse);
    end;
  end;
//     Close;
//     Open; }
  Result:=False;
end;

function TmCustomQuery.CheckFilter( Buffer: Pointer):boolean;
var
  SaveState: TDataSetState;
begin
   Result:=True;
   if not Assigned( FonFilter) then exit;
   SaveState := SetTempState(dsFilter);
   try
      FFilterBuffer := Buffer;
      FonFilter( Self, Result);
   finally
      RestoreState(SaveState);
   end;
end;

procedure TmCustomQuery.ClearCalcFields(Buffer: PChar);
begin
  FillChar(Buffer[RecordSize], CalcFieldsSize, 0);
end;

function TmCustomQuery.GetRecordSize: Word;
begin
  Result := FRecordSize;
end;

procedure TmCustomQuery.InternalHandleException;
begin
  Application.HandleException(Self)
end;

{ TmBlobStream }

constructor TmBlobStream.Create(Field: TBlobField; Mode: TBlobStreamMode);
begin
  FMode    := Mode;
  FField   := Field;
  FDataSet := FField.DataSet as TmCustomQuery;
  FFieldNo := FField.FieldNo;
  if not FDataSet.GetActiveRecBuf(FBuffer) then Exit;
{  if not FField.Modified then
  begin
    if Mode = bmRead then
    begin
      FCached := FDataSet.FCacheBlobs and (FBuffer = FDataSet.ActiveBuffer) and
        (FField.IsNull or (FDataSet.GetBlobData(FField, FBuffer) <> ''));
    end else
    begin
      FDataSet.SetBlobData(FField, FBuffer, '');
      if FField.ReadOnly then DatabaseErrorFmt(SFieldReadOnly, [FField.DisplayName]);
      if not (FDataSet.State in [dsEdit, dsInsert]) then DatabaseError(SNotEditing);
    end;
  end;  }
  FOpened := True;
  if Mode = bmWrite then Truncate;
end;

destructor TmBlobStream.Destroy;
begin
  if FOpened then
  begin
    if FModified then FField.Modified := True;
  end;
  if FModified then
  try
    FDataSet.DataEvent(deFieldChange, Longint(FField));
  except
    Application.HandleException(Self);
  end;
end;

function TmBlobStream.Read(var Buffer; Count: Longint): Longint;
begin
  Result := 0;
  if FOpened then
  begin
      if Count > Size - FPosition then
        Result := Size - FPosition else
        Result := Count;
      if Result > 0 then
      begin
        Move(PChar(FDataSet.GetBlobData(FField, FBuffer))[FPosition], Buffer, Result);
        Inc(FPosition, Result);
      end;
  end;
end;

function TmBlobStream.Write(const Buffer; Count: Longint): Longint;
var
//  Temp: Pointer;
  c:    AnsiString;
begin
  Result := 0;
  if FOpened then
  begin
{    if FField.Transliterate then
    begin
      GetMem(Temp, Count);
      c:= FDataSet.GetBlobData(FField, FBuffer);
      FDataSet.SetBlobData(FField, FBuffer, '');
      try
         SetLength( c, FPosition+Count);
         Move( Buffer, PChar(c)[FPosition], Count);
         FDataSet.SetBlobData(FField, FBuffer, c);

//         SetLength( PmBlobDataArray( FBuffer + DataSet.FBlobCacheOfs)[FField.Offset], FPosition+Count);
//         Move( Buffer, PChar(PmBlobDataArray( FBuffer + FBlobCacheOfs)[FField.Offset])[FPosition], Count);
//        AnsiToNativeBuf(FDataSet.Locale, @Buffer, Temp, Count);
//        Check(DbiPutBlob(FDataSet.Handle, FBuffer, FFieldNo, FPosition,
//          Count, Temp));
      finally
        FreeMem(Temp, Count);
      end;
    end else
//      Check(DbiPutBlob(FDataSet.Handle, FBuffer, FFieldNo, FPosition,
//        Count, @Buffer)); }
      c:= FDataSet.GetBlobData(FField, FBuffer);
      FDataSet.SetBlobData(FField, FBuffer, '');
      SetLength( c, FPosition+Count);
      Move( Buffer, PChar(c)[FPosition], Count);
      FDataSet.SetBlobData(FField, FBuffer, c);

    Inc(FPosition, Count);
    Result := Count;
    FModified := True;
//    FDataSet.SetBlobData(FField, FBuffer, '');
  end;
end;

function TmBlobStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  case Origin of
    0: FPosition := Offset;
    1: Inc(FPosition, Offset);
    2: FPosition := GetBlobSize + Offset;
  end;
  Result := FPosition;
end;

procedure TmBlobStream.Truncate;
begin
  if FOpened then
  begin
    FModified := True;
    FDataSet.SetBlobData(FField, FBuffer, '');
  end;
end;

function TmBlobStream.GetBlobSize: Longint;
begin
  Result := 0;
  if FOpened then
      Result := Length(FDataSet.GetBlobData(FField, FBuffer));
end;

end.
